#include "s2hat_cuda_common.cuh"
#include "s2hat_cuda_tools.cuh"

// File: s2hat_alm2map_cuda.cu
// Created by Ovidiu Hupca, Mikolaj Szydlarski, Giulio Fabbian on 2011-06-09
//  
// Copyright (C) 2012 Joel Falcou, Laura Grigori, Ovidiu Hupca, Radek Stompor, Mikolaj Szydlarski
// 
// This file is part of S2HAT_CUDA.
// 
// S2HAT_CUDA is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// S2HAT_CUDA is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with S2HAT_CUDA (see 'S2HAT_CUDA_LICENSE.txt' file). 
// If not, see <http://www.gnu.org/licenses/>.

#undef NO_OUT
#define SOFT_VERBOSE 0

__device__ __constant__ int nmvals_DCONST;
__device__ __constant__ int nlmax_DCONST;
__device__ __constant__ int nringsobs_DCONST;

typedef struct _deltam_arguments {

	int *mvals;

	double *cth_D;
	double *sth_D;
	double *local_mfac;
	
	cuDoubleComplex* local_alms;
	
	cuDoubleComplex* b_north; 
	cuDoubleComplex* b_south;

} deltam_arguments;

int static redistInterProductsAlm2Map(int nrings, int nmvals, int ndata, double complex* data, int myid, int numprocs, MPI_Comm comm) {

	int nringsall;

	// note: nmaps is assumed to be 1 and code for multiple maps is removed

	if ( MPI_Allreduce( &nrings, &nringsall, 1, MPI_INT, MPI_SUM, comm) ) perror("[!] redistInterProductsAlm2Map()::MPI_Allreduce ERROR [!]\n");

	int *mperproc = (int*) calloc(numprocs, sizeof(int)); 
	if (!mperproc)	{ perror("[mem error] Unable to allocate mperproc!\n"); return 1; }

	if ( MPI_Allgather(&nmvals, 1, MPI_INT, mperproc, 1, MPI_INT, comm) ) perror("[!] redistInterProductsAlm2Map()::MPI_Allgather ERROR [!]\n");

	int *ringsperproc = (int*) calloc(numprocs, sizeof(int)); 
	if (!ringsperproc)	{ perror("[mem error] Unable to allocate ringsperproc!\n"); return 1; }

	if ( MPI_Allgather( &nrings, 1, MPI_INT, ringsperproc, 1, MPI_INT, comm) ) perror("[!] redistInterProductsAlm2Map()::MPI_Allgather ERROR [!]\n");

	int *send_counts = (int*) calloc(numprocs, sizeof(int));
	int *recv_counts = (int*) calloc(numprocs, sizeof(int));
	if (!send_counts || !recv_counts)	{ perror("[mem error] Unable to allocate send_counts or recv_counts!\n"); return 1; }

	for (int i = 0; i < numprocs; ++i) {
		send_counts[i]  = nmvals*ringsperproc[i]; 
		recv_counts[i]  = nrings*mperproc[i];
	}

	int *send_disps = (int*) calloc(numprocs, sizeof(int));
	int *recv_disps = (int*) calloc(numprocs, sizeof(int));
	
	if (!send_disps || !recv_disps)	{ perror("[mem error] Unable to allocate send_disps or recv_disps!\n"); return 1; }

	recv_disps[0] = 0;
	send_disps[0] = 0;
      
	for(int i = 1; i < numprocs; ++i) {
		send_disps[i] = send_disps[i-1]+send_counts[i-1];
		recv_disps[i] = recv_disps[i-1]+recv_counts[i-1];
	}
	
	double complex* tmp = (double complex*) calloc(nmvals * nringsall, sizeof(double complex));
	if (!tmp) { perror("[mem error] Unable to allocate tmp!\n"); return 1; }
	
	// reshuffle to a ring-wise ordering here

	int ii = 0, jj = 0;
	size_t j = NULL, k = NULL;
	
	for (int i = 0; i < nmvals; ++i) {
		jj = 0;
		for (j = 0; j < numprocs; ++j) {
			jj += i * ringsperproc[j];
			for (k = 0; k < ringsperproc[j]; ++k) {
				tmp[jj+k] = data[ii];
				++ii;
			}
			jj += (nmvals-i)*ringsperproc[j];
		}
	}

/*
	MPI_Datatype MPI_DOUBLE_COMPLEXx;
	ierr = MPI_Type_contiguous(2, MPI_DOUBLE, &MPI_DOUBLE_COMPLEXx);	// datatype with 2 doubles
	ierr = MPI_Type_commit(&MPI_DOUBLE_COMPLEXx);
*/

	//problems might occur since MPI_DOUBLE_COMPLEX is actually a fortran type, not C-type...In this case, define e new datatype as above
	if (MPI_Alltoallv( tmp, send_counts, send_disps, MPI_DOUBLE_COMPLEX, data, recv_counts, recv_disps, MPI_DOUBLE_COMPLEX, comm )) perror("[!] redistInterProductsAlm2Map()::MPI_Alltoallv ERROR [!]\n");

	free(tmp);
	free(send_counts);
	free(send_disps);
	free(recv_counts);
	free(recv_disps);     
	free(ringsperproc);

	return 0;
} 

#ifdef CUFFT
	int static s2hat_ring_synthesis_cuda(int nlmax, int nmmax, double complex* datain, int nph, double* dataout, int nphmx, double kphi0, struct cufft_counters_c2r *cufftc, int iring) {
		
		const int ksign = +1;
		
		int im, jm, jmax, iw, m, k = nph/2;
		double complex* bw;
		double complex dat;
		double arg, dcosarg0, dcosarg1, dsinarg0, dsinarg1, tmp;
	  
	  double complex* fft_vec_in = cufftc->fft_vec_in;
	
	 	S2HATALLOC(bw, double complex, k+1);
		if (!bw) { perror("Unable to allocate bw!\n"); return 1; }
	
		// if (nmmax > nph) all frequencies [-m,m] are wrapped in [0,nph-1]
		// otherwise the missing high frequncies are set to zero
		
		memset(fft_vec_in, 0, (k+1)*sizeof(double complex)); 
	
		bw[0] = datain[0];
	
		dcosarg1 = 1E0;	// hope this is right <- dcosarg1 = 1.d0
		dsinarg1 = 0E0;
	
		arg = nph*kphi0;
		dcosarg0 = cos(arg);
		dsinarg0 = sin(arg);
	
		k = nmmax/nph;
	
		// computes only positive freqs as needed for FFT
	
		for ( im  = 0; im <= k; ++im) {
			jm = 0;
			m = im*nph + jm;
	
			if ((m <= nmmax) && (m > 0)) bw[jm] = bw[jm] + (2.0 * datain[m] * (  dcosarg1 + I*dsinarg1 ));
	
			jmax = min( nph/2, nmmax-im*nph);
	
			for( jm  = 1; jm<=jmax; ++jm) {
				m = im*nph + jm;
				bw[jm] = bw[jm] + datain[m] * (dcosarg1 + I*dsinarg1);					// #
			}
	
			tmp = dcosarg1*dcosarg0 + dsinarg1*dsinarg0;
			dsinarg1 = dcosarg1*dsinarg0 - dsinarg1*dcosarg0;
			dcosarg1 = tmp;
	
			jmax = min( nph-1, nmmax-im*nph);
	
			for(jm = (nph+1)/2; jm<=jmax; ++jm) {                      // and now negative one (-m)
				m = im*nph + jm;
				bw[ nph-jm ] = bw[ nph-jm ] + conj(datain[m])* (dcosarg1 - I*dsinarg1);			// #
			}
		}
	
		// and now account for the shift of the first pixel
	
		fft_vec_in[0] = bw[0];
	
		arg = ksign*kphi0;
		dcosarg0 = cos(arg);
		dsinarg0 = sin(arg);
	
		dcosarg1 = dcosarg0;
		dsinarg1 = dsinarg0;
	
		for ( iw = 1; iw <= (nph-1)/2; ++iw) {           // to allow for odd nph values
			dat  = bw[iw] * (dcosarg1 + I*dsinarg1);								// #
			tmp  = dcosarg1*dcosarg0-dsinarg1*dsinarg0;
			dsinarg1 = dcosarg1*dsinarg0+dsinarg1*dcosarg0;
			dcosarg1 = tmp;
			fft_vec_in[iw] = dat;	
		}
	
		// if nph is even then add the Nyquist mode now i.e., iw = nph/2
	
		if( 2*(nph/2) == nph ) {
			dat = bw[nph/2] * ( dcosarg1 +I*dsinarg1);							// #
			fft_vec_in[nph/2] = dat;
		}
		
		// alloc & copy datain to GPU //
	
		double complex* datain_D = NULL;
		double*        dataout_D = NULL;
	
		cudaError_t res2 = cudaMalloc( (void**) &datain_D, nph * sizeof(double complex) );
		cudaError_t res3 = cudaMalloc( (void**) &dataout_D, max(nphmx, 2*nph)*sizeof(double) );
	
		if ( (res2  != cudaSuccess) || (res3  != cudaSuccess) ) { 
			fprintf(stderr,"[e] --- Unable to allocate FFT data on device!\n"); return -1; 
		}
	
		res2 = cudaMemcpy (datain_D, fft_vec_in, (nph) * sizeof(double complex),    cudaMemcpyHostToDevice);
	
		if ( res2  != cudaSuccess ) { 
			fprintf(stderr,"[e] --- Unable to copy FFT data to device!\n"); 
			return -1; 
		}
	
		// plan and execute FFT //
	
		if( cufftc->cufft_c2r_plan_set != nph) {	// create a new plan or reuse?
			cufftDestroy(cufftc->cufft_c2r_plan);
		  
			cufftResult res = cufftPlan1d((cufftHandle*) &cufftc->cufft_c2r_plan, nph, CUFFT_Z2D, 1);
			Check_CUFFT_Error(res,"[e] --- Error at FFT plan creation ", iring);
			cufftc->cufft_c2r_plan_set = nph;
			cufftc->cufft_call_count++;
		}
	   
		cufftSetCompatibilityMode((cufftHandle) cufftc->cufft_c2r_plan, CUFFT_COMPATIBILITY_NATIVE);
		cufftResult res = cufftExecZ2D( (cufftHandle) cufftc->cufft_c2r_plan, (cufftDoubleComplex *) datain_D, (cufftDoubleReal*) dataout_D );
		Check_CUFFT_Error(res,"[e] --- Error at FFT execution ", iring);
	
		if (res == CUFFT_EXEC_FAILED) printf("Error in cuFFT \tnph     = %d\n \tnmmax    =%d\n", nph, nmmax);
		cufftc->cufft_full_count++;
	
		//copy fft result to dataout from GPU and free //
		res2 = cudaMemcpy (dataout, dataout_D, nphmx * sizeof(double),    cudaMemcpyDeviceToHost);
	
		if ( res2  != cudaSuccess ) { fprintf(stderr,"[e] --- Unable to copy FFT data to device!\n"); return -1; }
	   
		s2hat_free(bw);
		cudaFree(datain_D);
		cudaFree(dataout_D);
	  return 0;
	}
#endif

__global__ void plm_calc_double( deltam_arguments arg) {
	
	int c, scalem, scalel, par_lm;
	double lam_lm, lam_0, lam_1, lam_2;
	
	double log2val;
	
	double corfac;
	double lam_mm;
	
		
	unsigned int tid = threadIdx.x;
	unsigned int t   = (blockIdx.x * blockDim.x + threadIdx.x);

	#define RECFACRE(L,M) (        sqrt ( (double)  (4 * (L+1)*(L+1) - 1) / (double) ( (L+1)*(L+1)-M*M ) )  )
	#define RECFACIM(L,M) (       rsqrt ( (double)  (4 * (L+1)*(L+1) - 1) / (double) ( (L+1)*(L+1)-M*M ) )  )
	
	const double cth = (t < nringsobs_DCONST) ? arg.cth_D[t] : 1.0;
	const double sth = (t < nringsobs_DCONST) ? arg.sth_D[t] : 0.0;
		
	__shared__ double rescale_tab[ 2*SMAX+1 ];	// share the rescaling table with all the threads -> tradeoff memory consumption for increased shmem conflicts
	
	if (tid < SMAX) {
		rescale_tab[tid] = POWER(FL_LARGE, ((int)(tid) - SMAX));
	} else if (tid > SMAX && tid < RESCALE_TAB_SIZE ) {
		rescale_tab[tid] = POWER(FL_LARGE, (int)(tid % SMAX));
	} else if (tid == SMAX) {
		rescale_tab[tid] = 1.0;
	} 
	__syncthreads();
	
	const double ovflow = RESCALE_TAB_S(1);
	const double unflow = RESCALE_TAB_S(-1);
		
	// recfac variables //
	
	__shared__ int mvals_ck[ NR_THREADS_PER_BLOCK_A2M ];

	__shared__ double mfac_ck[ NR_THREADS_PER_BLOCK_A2M ];
	
	__shared__ double amm_re_ck[ NR_THREADS_PER_BLOCK_A2M ];
	__shared__ double amm_im_ck[ NR_THREADS_PER_BLOCK_A2M ];

	__shared__ double alm_re_ck[ NR_THREADS_PER_BLOCK_A2M ];
	__shared__ double alm_im_ck[ NR_THREADS_PER_BLOCK_A2M ];
	
	__shared__ double recfac_mm_ck[ NR_THREADS_PER_BLOCK_A2M ];
	
	__shared__ double recfac_re_ck[ RECFAC_CHUNK_SIZE_A2M +1 ];
	__shared__ double recfac_im_ck[ RECFAC_CHUNK_SIZE_A2M +1 ];
	
	PHX_DoubleComplexStruct b_ns; // finall structure to feel
	
	int mindx_ck = NR_THREADS_PER_BLOCK_A2M;
	
	for (size_t mindx_global = 0; mindx_global < nmvals_DCONST; mindx_global++, mindx_ck++) {
		
		if (mindx_ck == NR_THREADS_PER_BLOCK_A2M) {
			
			unsigned int m_tid = mindx_global + tid;
			
			__syncthreads();
			
			if (m_tid < nmvals_DCONST) { 
				
				mvals_ck[tid] = arg.mvals[m_tid];
				mfac_ck[tid]  = arg.local_mfac[m_tid];
				
				const int mv = mvals_ck[tid];
				
				cuDoubleComplex& cu_tmp = arg.local_alms[(m_tid)*(nlmax_DCONST+1) + mv];
				
				amm_re_ck[tid] = cuCreal(cu_tmp);
				amm_im_ck[tid] = cuCimag(cu_tmp);
				
				recfac_mm_ck[tid] = RECFACRE(mv,mv);
			}//~endIf
			
			__syncthreads();
			
			mindx_ck = 0; 
			
		}//~end if mindx_ck   
		
		const int m = mvals_ck[ mindx_ck ];
		const int l_min = fmax(m, (m - HPX_MXL0) / ( HPX_MXL1 * sth ) );
		const bool l_min_test = ( (nlmax_DCONST >= l_min) && (t < nringsobs_DCONST) );
		
		if (l_min_test) { // skip calculations when Ylm too small

			log2val = mfac_ck[mindx_ck] + m * log( sth ) * ALN2_INV;
			scalem = (int) (log2val / LOG2LG);
			
			corfac = POWER(FL_LARGE, max(scalem, RSMIN));
			lam_mm = POWER(2.0, (log2val - (scalem) * LOG2LG) );       		// rescaled lam_mm
			
			c = (m & 1) > 0; // condition result    
			lam_mm *= (-1)*(c) + (+1)*(1-c); // negative for odd m
			                                                
			// ---------- l = m ------------     //
			
			par_lm = 1;  // = (-1)^(l+m)
			c = (m >= l_min);
			
			lam_lm = corfac * lam_mm;
			
			b_ns.pos.re = ( (c)*lam_lm * amm_re_ck[mindx_ck] );
			b_ns.pos.im = ( (c)*lam_lm * amm_im_ck[mindx_ck] );
			
			b_ns.neg.re = 0;
			b_ns.neg.im = 0;
			
			// ---------- l > m ----------      //    
			
			lam_0  = 0.0;
			lam_1  = 1.0;
			scalel = 0;
			
			lam_2  = cth * lam_1 * recfac_mm_ck[mindx_ck];
		} //~end if 
		
		for (int l = m+1, k1 = RECFAC_CHUNK_SIZE_A2M, k2 = 1, ck_indx = NR_THREADS_PER_BLOCK_A2M; l <= nlmax_DCONST; l++, k1++, k2++, ck_indx++) {
			
			if (ck_indx == NR_THREADS_PER_BLOCK_A2M) {
				
				__syncthreads();
				
				unsigned int ck_tid = l + tid;
				
				if (ck_tid <= nlmax_DCONST) {
					cuDoubleComplex & cu_tmp = arg.local_alms[(mindx_global)*(nlmax_DCONST+1) + ck_tid];
					alm_re_ck[tid] = cuCreal(cu_tmp);
					alm_im_ck[tid] = cuCimag(cu_tmp);
				}
				
				ck_indx = 0;
				__syncthreads();
			}
			
			if ( k1 == RECFAC_CHUNK_SIZE_A2M ) {
				__syncthreads();
				#pragma unroll
				for (int x = tid, y = l-1 + tid; x <= RECFAC_CHUNK_SIZE_A2M; x+=NR_THREADS_PER_BLOCK_A2M, y+=NR_THREADS_PER_BLOCK_A2M) {
					const double aux_re = RECFACRE(y,m);
					const double aux_im = 1 / aux_re;	
					recfac_re_ck[x] = aux_re;
					recfac_im_ck[x] = aux_im;
				}
				// reset chunk indices
				k1=0;
				k2=1;
				__syncthreads();
			}
			
			if (l_min_test) {
				
				par_lm = - par_lm;  // = (-1)^(l+m)
				
				if (l >= l_min) {
					
					lam_lm = lam_2 * corfac * lam_mm;
					
					if (par_lm > 0) {
						b_ns.pos.re += lam_lm * alm_re_ck[ck_indx];
						b_ns.pos.im += lam_lm * alm_im_ck[ck_indx];
					} else {
						b_ns.neg.re += lam_lm * alm_re_ck[ck_indx];
						b_ns.neg.im += lam_lm * alm_im_ck[ck_indx];
					}
				}
				
				// two-point reccurence //
				
				lam_0 = lam_1 * recfac_im_ck[k1];
				lam_1 = lam_2;
				lam_2 = (cth * lam_1 - lam_0) * recfac_re_ck[k2];
				
				// overflow/underflow handling //
				
				const double abs_lam_2 = fabs(lam_2);
				
				if (abs_lam_2 > ovflow) {
					lam_1 = lam_1*unflow;
					lam_2 = lam_2*unflow;
					scalel= scalel + 1;
					
					const int rtx = max(scalel+scalem, RSMIN);
					
					corfac = RESCALE_TAB_S(rtx);
			
				} else if (abs_lam_2 < unflow) {
					lam_1 = lam_1*ovflow;
					lam_2 = lam_2*ovflow;
					scalel= scalel - 1;
					
					const int rtx = max(scalel+scalem, RSMIN);
					
					corfac = RESCALE_TAB_S(rtx);
				}//~end ELSE
			}
		} // loop on l                                             
		
		if (l_min_test) {
			
			#ifdef NO_OUT
				const int ith = t;   // global index 
			#else
				const int ith = mindx_global*(nringsobs_DCONST) + t;   // global index 
			#endif
			
			arg.b_north[ith] = make_cuDoubleComplex( b_ns.pos.re + b_ns.neg.re, b_ns.pos.im + b_ns.neg.im );
			arg.b_south[ith] = make_cuDoubleComplex( b_ns.pos.re - b_ns.neg.re, b_ns.pos.im - b_ns.neg.im );
			
		} // test on nlmax   	// 				par_lm = - par_lm;  // = (-1)^(l+m)
	}//~end loop over mindix
}  

#ifdef __cplusplus
	extern "C" {
#endif   

void s2hat_alm2map_cuda (
	int plms, // no use
	s2hat_pixeltype pixelization,
	s2hat_scandef scan,
	int nlmax,
	int nmmax,
	int nmvals,
	int* mvals, 
	int nmaps, // no use
	int nstokes, // no use 
	int first_ring, 
	int last_ring, 
	int map_size, 
	double *local_map, 
	int lda, 
	s2hat_dcomplex *local_alm,
	int nplm, // no use
	double* local_plm, // no use
	int numprocs,
	int myid,
	MPI_Comm comm)

{  
	
	if (plms > 0 ) {
		if (myid == 0) printf("[!] warning !! s2hat_alm2map_cuda( .. plms > 0 ..)  support only plms == 0 \n");
		plms = 0;
	}
	
	if (nmaps != 1) {
		perror("[!] s2hat_alm2map_cuda( .. nmaps != 1 ..) do not support multi-maps. Use s2hat_alm2map(..) instead!\n");
		exit(-1);
	} else if ( nstokes != 1 ) {
		perror("[!] s2hat_alm2map_cuda( .. nstokes != 1 ..) do not support spin transforms Use s2hat_alm2map(..) instead!\n");
		exit(-1);
	}
	
	
	// 0) Cuda stuff
	
	selectDevice(myid);
	
	#ifdef KERNEL_TIMER
		cudaEvent_t start, stop;
		cudaEventCreate(&start);
		cudaEventCreate(&stop);
	#endif
	
	double fft_start, fft_end, recurence_start, recurence_end, communication_start, communication_end, mem_sent_start, mem_sent_end, mem_rec_start, mem_rec_end;
	
	const int nringsobs = scan.nringsobs;
	
	#ifdef NO_OUT
		const int virtual_size = (nringsobs + ( NR_THREADS_PER_BLOCK_A2M - (nringsobs % NR_THREADS_PER_BLOCK_A2M )));
	#endif
	 
	const bool do_partial_map = (pixelization.nringsall != scan.nringsobs);
		
	long int istart_south, istart_north;   // map related
	int i;

	double *ring = NULL;
	double complex* phring = NULL;

	int *indx = NULL;
	double *nph0 = NULL, *kphi0 = NULL;
	int *keep_north = NULL, *keep_south =NULL;
	
	// CUDA algorithm specific variables .. host variables  

	double    *cth_D = NULL, *sth_D = NULL;
	int *mvals_D = NULL;

	double complex* local_alms_D = NULL;
	
	bool mfac_test;
	int mfac_lastm;
	double mfac_two_x_m;
	double mfac_tmp_val;
	double* local_mfac_H = NULL;
	double* local_mfac_D = NULL;
	
	s2hat_cuda_times * ct = NULL; 
	
	deltam_arguments arg;

	cudaError_t resA[20];
	
	double complex* b_north_H;
	double complex* b_south_H;
	
	double complex* b_north_D;
	double complex* b_south_D;
		
	// fftw variables //
	#ifdef CUFFT
		struct cufft_counters_c2r cufftc={0,0,0,NULL};
		S2HATALLOC(cufftc.fft_vec_in, double complex, ((pixelization.nphmx)/2) );
	#else
		struct fftw_counters_c2r fftwc = {0,0,0,0, NULL};
		init_fft_c2r(&fftwc, (pixelization.nphmx)/2);
	#endif

	const int nrings  = (last_ring-first_ring+1);
	const int nlen    = max_( nringsobs * nmvals, nrings * (nmmax+1) );
	
	#ifdef HARD_VERBOSE

		printf("[%d] !! \tnringsobs*nmvals     = %lld\n", myid, scan.nringsobs*nmvals);
		printf("[%d] !! \tnrings*(nmmax+1)     = %d\n", myid, nrings*(nmmax+1));
		printf("[%d] !! \tnlen                 = %d\n", myid, nlen);

		printRuntimeParameters();

		printf("\n[%d] --- s2hat_alm2map_cud() Basic Parameters \n\n", myid);
	
		printf("[%d]\tnlmax     = %d\n",   myid, nlmax);
		printf("[%d]\tnmvals    = %d\n",   myid, nmvals);
		printf("[%d]\tnphmx     = %lld\n", myid, pixelization.nphmx);
		printf("[%d]\tnlen      = %d\n",   myid, nlen);
		printf("[%d]\tnrings    = %d\n",   myid, nrings);
		printf("[%d]\tnringsobs = %lld\n", myid, scan.nringsobs);
		printf("[%d]\tnringsall = %lld\n", myid, pixelization.nringsall);

	#endif
	
	// memory allocation //
	
	#ifdef MAPPED_MEMORY
		
		cudaHostAlloc( (void **) &b_north_H, nlen * sizeof(double complex), cudaHostAllocMapped );
		cudaHostAlloc( (void **) &b_south_H, nlen * sizeof(double complex), cudaHostAllocMapped );
		
	#else
		
		b_north_H = (double complex*) malloc ( nlen * sizeof(double complex) );
		b_south_H = (double complex*) malloc ( nlen * sizeof(double complex) );

		if (! (b_north_H && b_south_H ) ) {
			perror("Unable to allocate host memory!\n");
			goto finalize;
		}
		
	#endif
   
	#ifdef MAPPED_MEMORY

		cudaHostGetDevicePointer((void **) &b_north_D, (void *) b_north_H, 0);
		cudaHostGetDevicePointer((void **) &b_south_D, (void *) b_south_H, 0);

		cudaMemset(b_north_D, 0, nlen * sizeof(double complex) );   // pad with zeros
		cudaMemset(b_south_D, 0, nlen * sizeof(double complex) );

	#else
	
		#ifdef NO_OUT
		
		cudaMalloc( (void**) &b_north_D, virtual_size * sizeof(double complex) );
		cudaMalloc( (void**) &b_south_D, virtual_size * sizeof(double complex) );

		#else

		cudaMalloc( (void**) &b_north_D, nlen * sizeof(double complex) );
		cudaMalloc( (void**) &b_south_D, nlen * sizeof(double complex) );
		
		#endif

		if (! (b_north_D && b_south_D) ) {
			perror("[e] --- Unable to allocate DEVICE memory!\n");
			goto allocError;
		}
		
		#ifdef NO_OUT
		
		cudaMemset(b_north_D, 0, virtual_size * sizeof(double complex) );   // pad with zeros
		cudaMemset(b_south_D, 0, virtual_size * sizeof(double complex) );
		
		#else
		
		cudaMemset(b_north_D, 0, nlen * sizeof(double complex) );   // pad with zeros
		cudaMemset(b_south_D, 0, nlen * sizeof(double complex) );
		
		#endif
		
	#endif
	
	local_mfac_H = (double*) malloc(nmvals*sizeof(double));
	
	if (mfac_test = (mvals[0] == 0) ? true : false) {
		local_mfac_H[0] = 1.0;
	}
		
	mfac_tmp_val = 1.0;
	mfac_lastm = 1;
		
	for (i = (mfac_test) ? 1 : 0; i < nmvals; ++i) {
		while (mfac_lastm <= mvals[i]) {
			mfac_two_x_m = 2.0 * mfac_lastm;
			mfac_tmp_val = mfac_tmp_val * sqrt((mfac_two_x_m + 1.0) / mfac_two_x_m);
			++mfac_lastm;
		}
		local_mfac_H[i] = mfac_tmp_val;
	}
		
	for (i = 0; i < nmvals; ++i) {
		local_mfac_H[i] = log(SQ4PI_INV * local_mfac_H[i]) * ALN2_INV;
	}
		
	resA[8] = cudaMalloc( (void**) &local_mfac_D, nmvals*sizeof(double));
	resA[9] = cudaMemcpy (local_mfac_D, local_mfac_H, nmvals*sizeof(double), cudaMemcpyHostToDevice);
		
	if ( (resA[8] || resA[9] ) != cudaSuccess ) {
		fprintf(stderr,"[e%d] --- Unable to allocate OR copy data from host to device (local_mfac_H -> local_mfac_D)!\n", myid);
		goto finalize;
	}
	
   memset(local_map, 0, map_size * sizeof(double) );  // set the whole map to zero 

	cudaMalloc( (void**) &local_alms_D, nmvals*(nlmax+1) * sizeof(double complex) );

	if (! local_alms_D ) {
		fprintf(stderr,"[e%d] --- Unable to allocate DEVICE memory for map data!\n", myid);

		Check_CUDA_Error(" .. ");	
	
		goto allocError;
	}
	
	if (do_partial_map) {
		cudaMalloc( (void**) &cth_D,  (nringsobs + ( NR_THREADS_PER_BLOCK_A2M - (nringsobs % NR_THREADS_PER_BLOCK_A2M ))) * sizeof(double) );
		cudaMalloc( (void**) &sth_D,  (nringsobs + ( NR_THREADS_PER_BLOCK_A2M - (nringsobs % NR_THREADS_PER_BLOCK_A2M ))) * sizeof(double) );
	} else {
		cudaMalloc( (void**) &cth_D,  (pixelization.nringsall + ( NR_THREADS_PER_BLOCK_A2M - (pixelization.nringsall % NR_THREADS_PER_BLOCK_A2M ))) * sizeof(double) );
		cudaMalloc( (void**) &sth_D,  (pixelization.nringsall + ( NR_THREADS_PER_BLOCK_A2M - (pixelization.nringsall % NR_THREADS_PER_BLOCK_A2M ))) * sizeof(double) );
	}

	cudaMalloc( (void**) &mvals_D, nmvals * sizeof(int) );
	
	if (! (cth_D && sth_D && mvals_D) ) {
		fprintf(stderr,"[e%d] --- Unable to allocate DEVICE memory for tranformation !\n", myid);
		goto allocError;
	}

	// getMemoryRequirement(nlen, nlmax, nmmax,nmvals, nringsall);

	// Copy host data to device //
	
	if (ct != NULL) mem_sent_start = MPI_Wtime();
	
	if (do_partial_map) {
		
		int scan_start = -1;
		
		for (int i = 0; i < pixelization.nringsall; ++i) {
			if (scan.fl[i] == 1) {
				scan_start = i;
				break;
			}
		}
		
		if (scan_start != -1) {
			resA[2] = cudaMemcpy (cth_D,  &pixelization.cth[scan_start],  nringsobs * sizeof(double), cudaMemcpyHostToDevice);
			resA[3] = cudaMemcpy (sth_D,  &pixelization.sth[scan_start],  nringsobs * sizeof(double), cudaMemcpyHostToDevice);
		} else {
			perror("scan_start == -1, there is no part of map to observe !!\n");
		}
		
	} else {
		resA[2] = cudaMemcpy (cth_D,  pixelization.cth,  pixelization.nringsall * sizeof(double),       cudaMemcpyHostToDevice);
		resA[3] = cudaMemcpy (sth_D,  pixelization.sth,  pixelization.nringsall * sizeof(double),       cudaMemcpyHostToDevice);
	}
	
	resA[7] = cudaMemcpy (mvals_D,   mvals,   nmvals * sizeof(int),       cudaMemcpyHostToDevice);

	if ( (resA[2] || resA[3] || resA[7] ) != cudaSuccess ) {
		fprintf(stderr,"[e%d] --- Unable to copy data from host to device!\n", myid);
		goto finalize;
	}
	
	resA[8] = cudaMemcpy (local_alms_D,   local_alm,   nmvals*(nlmax+1) * sizeof(double complex),    cudaMemcpyHostToDevice);
  
	if ( resA[8]  != cudaSuccess ) {
		fprintf(stderr,"[e%d] --- Unable to copy data from host to device!\n", myid);
		goto finalize;
	}
	
	cudaMemcpyToSymbol("nmvals_DCONST", &(nmvals), sizeof(int), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol("nlmax_DCONST",  &(nlmax),  sizeof(int), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol("nringsobs_DCONST", &(nringsobs), sizeof(int), 0, cudaMemcpyHostToDevice);
	
	if (ct != NULL) mem_sent_end = MPI_Wtime(); 

	Check_CUDA_Error("[e] --- Before kernel launch");
   
	#ifndef MFAC_ON_GPU
		arg.local_mfac = local_mfac_D;
	#endif

	arg.cth_D = cth_D;
	arg.sth_D = sth_D;  
	arg.mvals = mvals_D;
	arg.local_alms = (cuDoubleComplex*) local_alms_D;
	
	arg.b_north = (cuDoubleComplex*) b_north_D; 
	arg.b_south = (cuDoubleComplex*) b_south_D;
	
  // --- only works on >= Fermi or Tesla

	#ifdef PREFER_L1
		cudaFuncSetCacheConfig(plm_calc_double, cudaFuncCachePreferL1);
	#else
		cudaFuncSetCacheConfig(plm_calc_double, cudaFuncCachePreferShared);
	#endif  
	
	Check_CUDA_Error("cudaFuncSetCacheConfig error");  

	if (ct != NULL) recurence_start = MPI_Wtime();
	
	if ( do_partial_map ) {
		plm_calc_double<<< ( (nringsobs + ( NR_THREADS_PER_BLOCK_A2M - (nringsobs % NR_THREADS_PER_BLOCK_A2M ))) / NR_THREADS_PER_BLOCK_A2M), NR_THREADS_PER_BLOCK_A2M >>>(arg); 
		cudaThreadSynchronize(); 
	} else {
		plm_calc_double<<< ( (pixelization.nringsall + ( NR_THREADS_PER_BLOCK_A2M - (pixelization.nringsall % NR_THREADS_PER_BLOCK_A2M ))) / NR_THREADS_PER_BLOCK_A2M), NR_THREADS_PER_BLOCK_A2M >>>(arg); 
		cudaThreadSynchronize();
	}
	
	if (ct != NULL) recurence_end = MPI_Wtime();
  
 	Check_CUDA_Error("[e] --- Kernel error");    

	if (SOFT_VERBOSE) printf("\n[%d] --- End computation on GPU                              --- [!]\n",myid);
	
	if (ct != NULL) mem_rec_start = MPI_Wtime();

	#ifndef MAPPED_MEMORY
		#ifndef NO_OUT
		resA[0] = cudaMemcpy (b_north_H, b_north_D, nlen * sizeof(double complex), cudaMemcpyDeviceToHost);
		Check_CUDA_Error("[e] --- Memcpy (D->H) b_north"); 
		
		resA[1] = cudaMemcpy (b_south_H, b_south_D, nlen * sizeof(double complex), cudaMemcpyDeviceToHost);
		Check_CUDA_Error("[e] --- Memcpy (D->H) b_south");
		#endif
	#endif  
	
	if (ct != NULL) mem_rec_end = MPI_Wtime();

	if ( (resA[0] || resA[1]  || resA[2]   || resA[3] ) != cudaSuccess ) {
		fprintf(stderr,"[e] --- Unable to copy data from device to host!\n");
		goto finalize;
	}
   
	if (SOFT_VERBOSE) printf("[%d] --- Ok Ok .. GPU return with data !!",myid);
	
	nph0 = (double*) calloc( nrings , sizeof(double) );	
	kphi0= (double*) calloc( nrings , sizeof(double) );
	keep_north = (int*) calloc( nrings , sizeof(int) );	
	keep_south = (int*) calloc( nrings , sizeof(int) );
	
	if (!(nph0 && kphi0 && keep_north && keep_south)) {
			perror("[e] --- Unable to allocate phi memory!\n");
			goto finalize;
	}
	
	for (i = first_ring; i<=last_ring; ++i) {	//TODO: remove or at least improve this step (memset)
		nph0 [i-first_ring]      = (double) pixelization.nph[i-1];
		kphi0[i-first_ring]      = (double) pixelization.kphi[i-1];
		keep_north[i-first_ring] = (int) scan.nfl[i-1];
		keep_south[i-first_ring] = (int) scan.sfl[i-1];
	}
	
	// :: -- redistribution
	
	if (SOFT_VERBOSE && myid == 0) printf("\n[!] --- Redistribution...\n");
	
	if (ct != NULL) communication_start = MPI_Wtime();
   
	redistInterProductsAlm2Map( nrings, nmvals, nlen, b_north_H, myid, numprocs, comm);
	redistInterProductsAlm2Map( nrings, nmvals, nlen, b_south_H, myid, numprocs, comm);
	
	if (ct != NULL) communication_end = MPI_Wtime();
   
	if (SOFT_VERBOSE && myid == 0) printf("\n[!] --- Redistribution... DONE\n\n");
	
	ring      = (double*)   		calloc(pixelization.nphmx, sizeof(double));	
	phring    = (double complex* )  calloc((nmmax+1),sizeof(double complex));
	indx      = (int*)		calloc((nmmax+1),sizeof(int));	
	
	if (!(ring && phring && indx)) {
		perror("[e] --- Unable to allocate ring memory!\n");
		goto finalize;
	}
	
	fix_mmap( 1, nmmax, indx, nmvals, mvals, myid, numprocs, comm);
	
	if (SOFT_VERBOSE) printf("\n[!] --- Starting map gen... with nrings = %d, FFT by ", nrings);
	
	#ifdef CUFFT
		if (SOFT_VERBOSE) printf("CUFFT");
	#else
		if (SOFT_VERBOSE) printf("FFTW");
	#endif
	
	istart_north = 0;
	istart_south = map_size;
	
	if (ct != NULL) fft_start = MPI_Wtime();
	
	for (int iring = 0; iring < nrings; ++iring) {
		const int nphl = nph0[iring];
		int ij;
	
		if ( keep_north[iring] == 1 ) {
			
			//  ---------------------------------------------------------------  //
	      //  sum_m  b(m,theta)*exp(i*m*phi)   -> f(phi,theta) by FFT         //
			//  --------------------------------------------------------------- //
			
			for (i = 0; i <= nmmax; ++i) {
				ij = iring + indx[i] * nrings;
				phring[i] = b_north_H[ij];
			}
	
			#ifdef CUFFT
				s2hat_ring_synthesis_cuda(nlmax, nmmax, phring, nphl, ring, pixelization.nphmx, kphi0[iring], &cufftc, iring);   	// CUDA FFT - north			
			#else
				s2hat_ring_synthesis(nlmax, nmmax, phring, nphl, ring, kphi0[iring], &fftwc);   	// north hemisph. + equator 
			#endif
			
			memcpy( local_map + istart_north, ring, nphl * sizeof(double)); // copy ring to final map		
		}
	
		istart_north = istart_north + nphl;
		istart_south = istart_south - nphl;
	
		if ( keep_south[iring] == 1 ) {
						
			for (i=0; i<=nmmax; i++) {
				ij = iring + indx[i] * nrings;
				phring[i] = b_south_H[ij];
			}
			#ifdef CUFFT
				s2hat_ring_synthesis_cuda(nlmax, nmmax, phring, nphl, ring, pixelization.nphmx, kphi0[iring], &cufftc, iring);	// CUDA FFT - south
			#else
				s2hat_ring_synthesis(nlmax, nmmax, phring, nphl, ring, kphi0[iring], &fftwc);		// south hemisph. w/o equat
			#endif
	
			memcpy( local_map + istart_south, ring, nphl * sizeof(double)); // copy ring to final map
		}
	}
	
	if (ct != NULL) fft_end = MPI_Wtime();
	
	if (ct != NULL) {
		ct->recurrence_dt = (recurence_end - recurence_start);
		ct->fft_dt        = (fft_end - fft_start);
		ct->comm_dt       = (communication_end - communication_start);
		ct->mem_sent_dt   = (mem_sent_end - mem_sent_start);
		ct->mem_recv_dt   = (mem_rec_end - mem_rec_start);
	}

allocError:
finalize:
	
	if (b_north_D != NULL) cudaFree(b_north_D);
	if (b_south_D != NULL) cudaFree(b_south_D);
	
	free(nph0);
	free(kphi0);
	free(keep_north);
	free(keep_south);
	
	free(ring);
	free(phring);
	free(indx);
	
	s2hat_free(b_north_H);
	s2hat_free(b_south_H);
	
	cudaFree(cth_D);
	cudaFree(sth_D);
	
	#ifdef CUFFT
		s2hat_free(cufftc.fft_vec_in);
	#else
		free_fft_c2r(&fftwc);
	#endif
}

#ifdef __cplusplus
	}
#endif  






