#include <mpi.h>
#include "s2hat.h"
#include "s2hat_utils.h"

// File: s2hat_example_alm2map2alm.c
// 
// Copyright (C) 2012 Joel Falcou, Laura Grigori, Ovidiu Hupca, Radek Stompor, Mikolaj Szydlarski
// 
// This file is part of S2HAT_CUDA.
// 
// S2HAT_CUDA is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// S2HAT_CUDA is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with S2HAT_CUDA (see 'S2HAT_CUDA_LICENSE.txt' file). 
// If not, see <http://www.gnu.org/licenses/>.

/* this code computes first alm2map transform from randomly drawn alm coefficients. Then it performs map2alm *
 * transform of the map and compares new alm to the randomly generated input. It collects  alm coefficients  *
 * and map and writes them to disc as a binary file if required. By default transforms are computed on the   *
 * full sky. It computes the power spectrum of the map from precomputed alm coefficients if asked.			 *.
 * Note that though in principle the code performs the 'full loop' almw2map followed by map2map the output   * 
 * alm coefficients will not agree perfectly with the input due to the bandwidth issues.                     *
 *                                                                                                           *
 *                                                                                      - rs@apc, 2010/02/10 *
 * Works correctly for a multiple maps and any accepted number of Stokes parameters                          *
 *                                                                                                           *
 *                                                                                      - rs@apc, 2010/09/07 *
 *                                                                                                           *
 * New routines added, now we can evaluate map2alm and alm2map using multi-threaded ('_mt' suffix) or cuda   *
 * ('_cuda' suffix) version. Works only for singular map                                                     *
 *                                                                                                           *
 *                                                                                      - ms@lri, 2011/09/20 *
 * Minor I/O edits and more command line options added														 *
 *                                                                                      - gf@apc, 2012/05/10 *
 *                                                                                                           */

static void s2hat_measure_errors_and_time (s2hat_dcomplex* alm_init, s2hat_dcomplex* alm_final, int nalms, double dt_alm2map, double dt_map2alm, int lmax, int nside, int proces_id) {
	int i;
	double sum=0, sum2=0, sum_final=0, global_sum_final=0, maxdiff=0, global_maxdiff=0;
	double dt_alm2map_min, dt_alm2map_max, dt_map2alm_min, dt_map2alm_max;
	
	for (i = 0; i < nalms; ++i) {
		double x = alm_final[i].re - alm_init[i].re;
		double y = alm_final[i].im - alm_init[i].im;
		sum += x*x + y*y;
		sum2 += alm_init[i].re*alm_init[i].re+alm_init[i].im*alm_init[i].im;
		
		if (fabs(x)>maxdiff) maxdiff = fabs(x);
		if (fabs(y)>maxdiff) maxdiff = fabs(y);
	}
	
	sum  = sqrt( sum / nalms );
	sum2 = sqrt(sum2 / nalms );
	
	sum_final = sum/sum2;
	
	
	MPI_Reduce( &sum_final, &global_sum_final, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	MPI_Reduce( &maxdiff, &global_maxdiff, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	
	MPI_Reduce( &dt_alm2map, &dt_alm2map_min, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
	MPI_Reduce( &dt_map2alm, &dt_map2alm_min, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
	
	MPI_Reduce( &dt_alm2map, &dt_alm2map_max, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	MPI_Reduce( &dt_map2alm, &dt_map2alm_max, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	
	
	if (proces_id == 0) {
		printf("[%d][s2hat, nside = %d,lmax = %d] Errors  <rms, maxerr> %e\t%e\n", proces_id, nside, lmax, global_sum_final, global_maxdiff);
		printf("[%d][s2hat, nside = %d,lmax = %d] alm2map_<min, avg, max>  %6.3f\t%6.3f\t%6.3f\t[s]\n", proces_id, nside, lmax,  dt_alm2map_min, (dt_alm2map_min + dt_alm2map_max)/2.0, dt_alm2map_max );
		printf("[%d][s2hat, nside = %d,lmax = %d] map2alm_<min, avg, max>  %6.3f\t%6.3f\t%6.3f\t[s]\n", proces_id, nside, lmax, dt_map2alm_min, (dt_map2alm_min + dt_map2alm_max)/2.0, dt_map2alm_max );
	}
}

s2hat_int4 main(s2hat_int4 argc, char * argv[]) {
	
	s2hat_int4 root  = 0;
	s2hat_int4 nmaps = 1  /* In the current version MT and CUDA routines only support nmaps=1  */; 
	
	s2hat_int4 myrank, nprocs;
	
	/* MPI initialization */
	
	MPI_Init( &argc, &argv);
	
	MPI_Comm_rank( MPI_COMM_WORLD, &myrank);
	MPI_Comm_size( MPI_COMM_WORLD, &nprocs);
	
	double map2alm_start, map2alm_end, alm2map_start, alm2map_end;
	double dt_map2alm, dt_alm2map;
	double dt_map2alm_min, dt_map2alm_max, dt_alm2map_min, dt_alm2map_max;
	
	s2hat_int4 nside;
	s2hat_int4 lmax;
	s2hat_int4 nrings;
	s2hat_int4 npix;
	s2hat_int4 pixchoice;
	s2hat_pixparameters pixpar; 
	
	if (argc < 4 || argc > 5) {
		printf("usage: ./s2hat_c_example_a2m2a.exe < glcp | ecp | healpix | glesp > <resolution_parameter> <lmax> <OPTIONAL: write_output>\n"
				"GLCP\t resolution_parameter: nrings (DEFAULT n pixel per ring is set to 2nrings)\n"
				"ECP\t resolution_parameter: nrings (DEFAULT n pixel per ring is set to 2nrings)\n"
				"HEALPix\t resolution_parameter: nside\n"
				"GLESP\t resolution_parameter: nrings (DEFAULT n pixel per ring is set to nrings)\n");
		exit(-1);
	}
	
	nside = atoi(argv[2]);
	lmax  = atoi(argv[3]);
	
	if (strcmp(argv[1],"glcp") == 0) {
		pixchoice = PIXCHOICE_GLCP;
		nrings = nside;
		pixpar.par1 = nrings; 
		pixpar.par2 = 2*nrings;
	} else if (strcmp(argv[1],"ecp") == 0) {
		pixchoice = PIXCHOICE_ECP;
		nrings = nside;
		pixpar.par1 = nrings; 
		pixpar.par2 = 2*nrings;
	} else if (strcmp(argv[1],"healpix") == 0) {
		pixchoice = PIXCHOICE_HEALPIX;
		pixpar.par1 = nside; 
		pixpar.par2 = 0;
	} else if (strcmp(argv[1],"glesp") == 0) {
		pixchoice = PIXCHOICE_GLESP;
		nrings = nside;
		pixpar.par1 = nrings; 
		pixpar.par2 = nrings;	
	} else {
		perror("Pick a correct pixelization scheme !!\n");
	}
			
	
	s2hat_int4 nlmax = lmax , nmmax = lmax;
	s2hat_int4 nstokes = 1;  /* no of Stokes params 1, 2 or 3 - In the current version MT and CUDA routines only support nstokes=1 */
	s2hat_int4 nspec = nstokes*(nstokes+1)/2;    /* no of map spectra to be computed - check the rules in the manual.*/
	s2hat_int4 plms = 0;     /* i.e., do not assume that plm are precomputed */
	s2hat_int4 nmvals, first_ring, last_ring, map_size;
	s2hat_int4 *mvals;
	s2hat_int4 one = 1, zero = 0;
	s2hat_int4 i, j, lout, mout;
	s2hat_int4 nalms;
	s2hat_int4 write_to_file = 0;
	
	s2hat_int8 nplm;
	
	s2hat_flt8 *local_map, *map;
	s2hat_flt8 *local_w8ring;
	s2hat_flt8 *cls;
	
	FILE *fout;
	
	s2hat_dcomplex *local_alm, *local_alm_final, *alms;
	
	s2hat_pixeltype cpixelization;
	s2hat_scandef cscan;
	
	s2hat_flt8 zbounds[2];
	
	if(argc == 5) write_to_file=atoi(argv[4]);
	
	/* - define the pixelization structure -> cpixelization */
	set_pixelization( pixchoice, pixpar, &cpixelization); 
	
	npix=cpixelization.npixsall;
	if (myrank == root) printf("\n[!] Parameters:\t pixtype = %s\t resolution_parameters = [ %d , %d ]\tlmax = %d\t npix = %d\n\n", argv[1] ,pixpar.par1,pixpar.par2, lmax, npix);
	
	
	/* - define the scan structure -> cscan : corresponding to the observed sky given by zbounds[0] >= cos theta >= zbounds[1] */
	
	zbounds[0] = 0.0; zbounds[1] = 0.0;
	// zbounds[0] = -0.258805; zbounds[1] = 0.258805; // uncomment if you want to use partial map feature
	zbounds2scan( zbounds, cpixelization, &cscan);
	
	/* computes the data distribution over the processors */
	
	get_local_data_sizes( plms, cpixelization, cscan, nlmax, nmmax, myrank, nprocs, &nmvals, &first_ring, &last_ring, &map_size, &nplm, root, MPI_COMM_WORLD);  
	
	/* number of equilatitude rings on this processor */
	nrings = last_ring-first_ring+1;
	
	/* m-values to be stored on this proc */
	S2HATALLOC(mvals, int, nmvals);
	
	find_mvalues( myrank, nprocs, nmmax, nmvals, mvals);
	
	/* number of alm coefficients stored on this proc */
	nalms = (nlmax+1)*nmvals*nstokes;

	S2HATALLOC(local_alm, s2hat_dcomplex, nalms);
	S2HATALLOC(local_map, s2hat_flt8, nmaps*map_size*nstokes);

	s2hat_random_alm(local_alm, mvals, nmvals, nlmax);
	
	/* calculate the alm coefficients */
	
	MPI_Barrier(MPI_COMM_WORLD);	 
	
	alm2map_start = MPI_Wtime();
    
	#ifdef S2HAT_MT
	  if (myrank == root) printf("\n[!] s2hat_alm2map_mt( ... ) ...................... ::\n");
	  s2hat_alm2map_mt( plms, cpixelization, cscan, nlmax, nmmax, nmvals, mvals, nmaps, nstokes, first_ring, last_ring,map_size, local_map, nlmax, local_alm, nplm, NULL, nprocs, myrank, MPI_COMM_WORLD );
	  if (myrank == root) printf("[!] s2hat_alm2map_mt( ... ) ............. DONE ... ::\n\n");
	#elif S2HAT_CUDA
	   if (myrank == root) printf("\n[!] s2hat_alm2map_cuda( ... ) .................... ::\n");
	   s2hat_alm2map_cuda( plms, cpixelization, cscan, nlmax, nmmax, nmvals, mvals, nmaps, nstokes, first_ring, last_ring, map_size, local_map, nlmax, local_alm, nplm, NULL, nprocs, myrank, MPI_COMM_WORLD );
	   if (myrank == root) printf("[!] s2hat_alm2map_cuda( ... ) ........... DONE ... ::\n\n");
	#else
	  if (myrank == root) printf("\n[!] s2hat_alm2map( ... ) ......................... ::\n"); 
	  s2hat_alm2map( plms, cpixelization, cscan, nlmax, nmmax, nmvals, mvals, nmaps, nstokes, first_ring, last_ring, map_size, local_map, nlmax, local_alm, nplm, NULL, nprocs, myrank, MPI_COMM_WORLD );
	  if (myrank == root) printf("[!] s2hat_alm2map( ... ) ................ DONE ... ::\n\n"); 
	#endif
	
	alm2map_end = MPI_Wtime();
	
	if (write_to_file) {
		
		if( myrank == root) { /* gather complete map in the memory of the proc root */
			S2HATALLOC(map, s2hat_flt8, nmaps*nstokes*npix);
		}
		
		for( i = 0; i < nmaps; ++i) { /* do the loop over maps */ 
			collect_map(cpixelization,nmaps,i,nstokes,&map[i*nstokes*npix],first_ring,last_ring,map_size,local_map,myrank,nprocs,root,MPI_COMM_WORLD);
		}//~endFor

		if(myrank ==root && write_to_file) {
			fout = fopen( "test_map.bin", "w"); fwrite( map, sizeof( double), nmaps*nstokes*npix, fout); fclose( fout);
		}
		
		if (myrank == root) s2hat_free(map);
	}
	
	S2HATALLOC(local_w8ring, s2hat_flt8, nrings*nstokes);
	
	for( i=0; i<nrings*nstokes; ++i) local_w8ring[i] = 1.0;
	
	S2HATALLOC(local_alm_final, s2hat_dcomplex, nalms);
	
	map2alm_start = MPI_Wtime(); 
	
	#ifdef S2HAT_MT
		if (myrank == root) printf("\n[!] s2hat_map2alm_mt( ... ) ...................... ::\n");
		s2hat_map2alm_mt( plms, cpixelization, cscan, nlmax, nmmax, nmvals, mvals, nmaps, nstokes, first_ring, last_ring, local_w8ring, map_size, local_map, nlmax, local_alm_final, nplm, NULL, nprocs, myrank, MPI_COMM_WORLD);
		if (myrank == root) printf("[!] s2hat_map2alm_mt( ... ) ............. DONE ... ::\n\n");
	#elif S2HAT_CUDA
		if (myrank == root) printf("\n[!] s2hat_map2alm_cuda( ... ) .................... ::\n");
		s2hat_map2alm_cuda( plms, cpixelization, cscan, nlmax, nmmax, nmvals, mvals, nmaps, nstokes, first_ring, last_ring, local_w8ring, map_size, local_map, nlmax, local_alm_final, nplm, NULL, nprocs, myrank, MPI_COMM_WORLD);
		if (myrank == root) printf("[!] s2hat_map2alm_cuda( ... ) ........... DONE ... ::\n\n");
	#else
		if (myrank == root) printf("\n[!] s2hat_map2alm( ... ) ......................... ::\n");
		s2hat_map2alm( plms, cpixelization, cscan, nlmax, nmmax, nmvals, mvals, nmaps, nstokes, first_ring, last_ring, local_w8ring, map_size, local_map, nlmax, local_alm_final, nplm, NULL, nprocs, myrank, MPI_COMM_WORLD);
		if (myrank == root) printf("[!] s2hat_map2alm( ... ) ................ DONE ... ::\n\n");
	#endif
	
	map2alm_end = MPI_Wtime();
		
	dt_alm2map = (alm2map_end - alm2map_start);
	dt_map2alm = (map2alm_end - map2alm_start);
	
	s2hat_measure_errors_and_time (local_alm, local_alm_final, nalms, dt_alm2map, dt_map2alm, lmax, nside, myrank);
	
	s2hat_free(local_w8ring);
	
	/* gather all computed alm in the memory of the proc 'root' */
	
	if (write_to_file) {
		if( myrank == root) {
			S2HATALLOC(alms, s2hat_dcomplex, nstokes*(nlmax+1)*(nmmax+1) );
		}
		for( i=0; i < nmaps; ++i) { /* do a loop over maps */
			collect_alms( nlmax, nmmax, nmaps, i, nstokes, nmvals, mvals, nlmax, local_alm, alms, myrank, nprocs, root, MPI_COMM_WORLD);
			
			if (myrank == root && write_to_file) {
				fout = fopen( "test_alm.bin", "w"); fclose( fout);
				fout = fopen( "test_alm.bin", "a"); fwrite( alms, sizeof( s2hat_dcomplex), nstokes*(nmmax+1)*(nlmax+1), fout); fclose( fout);
			}
		}//~endFor
		
		if(myrank == root) s2hat_free(alms);
	}
	
	/* go back from alms to map */
	
	s2hat_free(local_map);
	
	/* and compute (auto)spectra just for the first map */
	
	if(write_to_file) {
		if(myrank == root){
			S2HATALLOC(cls, s2hat_flt8, (nlmax+1)*nspec );
		}
		
		collect_cls( nmaps, zero, nstokes, nlmax, nmvals, mvals, nlmax, local_alm, nspec, cls, myrank, nprocs, root, MPI_COMM_WORLD);
		
		s2hat_free(local_alm);
		
		if(myrank ==root && write_to_file) {
			fout = fopen( "test_cl.bin", "w"); fwrite( cls, sizeof( double), (nlmax+1)*nspec, fout); fclose( fout);
		}
		
		s2hat_free(cls);
	}
	
	/* conclude */
	destroy_scan(cscan);
	destroy_pixelization( cpixelization);
	
	/* and finish up */
	MPI_Finalize();
  return(0);
}

