#include "s2hat_c_common.h"
#include <omp.h>

//  File: s2hat_map2alm_mt.c
//  Created by Mikolaj Szydlarski (mikolaj.szydlarski@inria.fr) on 2011-06-09
// 
// Copyright (C) 2012 Joel Falcou, Laura Grigori, Ovidiu Hupca, Radek Stompor, Mikolaj Szydlarski
// 
// This file is part of S2HAT_CUDA.
// 
// S2HAT_CUDA is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// S2HAT_CUDA is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with S2HAT_CUDA (see 'S2HAT_CUDA_LICENSE.txt' file). 
// If not, see <http://www.gnu.org/licenses/>.

void static gatherRingInfo(int first_ring, int nrings, int* keep_it, int nringsall, int* keep_it_all, int myid, int numprocs, MPI_Comm comm) {
	int i = 0;
	int* frings = NULL;
	int* ringsperproc = NULL;
	
	S2HATALLOC(ringsperproc, int, numprocs);
	
	MPI_Allgather(&nrings, 1, MPI_INT, ringsperproc, 1, MPI_INT, comm);
	
	S2HATALLOC(frings, int, numprocs);

	i = first_ring - 1;
	
	MPI_Allgather(&i, 1, MPI_INT, frings, 1, MPI_INT, comm);
	MPI_Allgatherv(keep_it, nrings, MPI_INT, keep_it_all, ringsperproc, frings, MPI_INT, comm);
	
	s2hat_free(ringsperproc);
	s2hat_free(frings);
}

int static redistInterProductsMap2Alm(int nrings, int nmvals, int ndata, double complex* data, int myid, int numprocs, MPI_Comm comm) {
	
	int i = 0;  
	int nringsall;
	double complex* tmp = NULL;

	// note: nmaps is assumed to be 1 and code for multiple maps is removed

	if (MPI_Allreduce(&nrings, &nringsall, 1, MPI_INT, MPI_SUM, comm)) printf("[!] redistInterProductsAlm2Map()::MPI_Allreduce ERROR [!]\n");

	int *mperproc = (int*) calloc(numprocs, sizeof (int));
	if (!mperproc) {
		printf("[mem error] Unable to allocate mperproc!\n");
		return 1;
	}

	if (MPI_Allgather(&nmvals, 1, MPI_INT, mperproc, 1, MPI_INT, comm)) printf("[!] redistInterProductsAlm2Map()::MPI_Allgather ERROR [!]\n");

	int *ringsperproc = (int*) calloc(numprocs, sizeof (int));
	if (!ringsperproc) {
		printf("[mem error] Unable to allocate ringsperproc!\n");
		return 1;
	}

	if (MPI_Allgather(&nrings, 1, MPI_INT, ringsperproc, 1, MPI_INT, comm)) printf("[!] redistInterProductsAlm2Map()::MPI_Allgather ERROR [!]\n");

	int *send_counts = (int*) calloc(numprocs, sizeof (int));
	int *recv_counts = (int*) calloc(numprocs, sizeof (int));
	
	if (!send_counts || !recv_counts) {
		printf("[mem error] Unable to allocate send_counts or recv_counts!\n");
		return 1;
	}

	for (i = 0; i < numprocs; ++i) {
	  	send_counts[i] = (nrings * mperproc[i]);
		recv_counts[i] = (nmvals * ringsperproc[i]);
	}

	free(mperproc);

	int *send_disps = (int*) calloc(numprocs, sizeof (int));
	int *recv_disps = (int*) calloc(numprocs, sizeof (int));

	if (!send_disps || !recv_disps) {
		printf("[mem error] Unable to allocate send_disps or recv_disps!\n");
		return 1;
	}

	recv_disps[0] = 0;
	send_disps[0] = 0;

	for (i = 1; i < numprocs; ++i) {
		send_disps[i] = send_disps[i - 1] + send_counts[i - 1];
		recv_disps[i] = recv_disps[i - 1] + recv_counts[i - 1];
	}


	S2HATALLOC(tmp, double complex, nmvals*nringsall );

	if (!tmp) {
		printf("[mem error] Unable to allocate tmp!\n");
		return 1;
	}
	
	if (MPI_Alltoallv(data, send_counts, send_disps, MPI_DOUBLE_COMPLEX, tmp, recv_counts, recv_disps, MPI_DOUBLE_COMPLEX, comm)) printf("[!] redistInterProductsMap2Alm()::MPI_Alltoallv ERROR [!]\n");

	// reshuffle to a ring-wise ordering here
	
	int ii = 0, jj = 0;
	int j = 0, k = 0;

	for (i = 0; i < nmvals; ++i) {
		jj = 0;
		for (j = 0; j < numprocs; ++j) {
			jj += (i * ringsperproc[j]);
			for (k = 0; k < ringsperproc[j]; ++k) {
				data[ii] = tmp[jj + k];
				++ii;
			}
			jj += (nmvals-i) * ringsperproc[j];
		}
	}

	s2hat_free(tmp);
	free(send_counts);
	free(send_disps);
	free(recv_counts);
	free(recv_disps);
	free(ringsperproc);

	return 0;
}

void s2hat_map2alm_mt(
	int precompute_plms,
	s2hat_pixeltype pixelization,
	s2hat_scandef scan,
	int nlmax,
	int nmmax,
	int nmvals,
	int* mvals,
	int nmaps,
	int nstokes,
	int first_ring,
	int last_ring,
	s2hat_flt8 *local_w8ring,
	int map_size,
	s2hat_flt8 *local_map,
	int lda,
	s2hat_dcomplex *local_alm,
	int64_t nplm,
	s2hat_flt8* local_plm,
	int numprocs,
	int myid,
	MPI_Comm comm ) {
		
		if (precompute_plms > 0) {
			if (myid == 0) printf("[!] warning !! s2hat_map2alm_mt( .. plms > 0 ..)  support only plms == 0\n");
			precompute_plms = 0;
		}
	
		if (nmaps != 1) {
			perror("[!] s2hat_map2alm_mt( .. nmaps != 1 ..) do not support multi-maps. Use s2hat_alm2map(..) instead!\n");
			exit(-1);
		} else if ( nstokes != 1 ) {
			perror("[!] s2hat_map2alm_mt( .. nstokes != 1 ..) do not support spin transforms Use s2hat_alm2map(..) instead!\n");
			exit(-1);
		}
	
	double fft_start, fft_end, recurence_start, recurence_end, communication_start, communication_end;
	
	const int nphmx = pixelization.nphmx;
	const int nrings   = (last_ring - first_ring + 1);
	const int nringsall = pixelization.nringsall;
	const int nringsobs = scan.nringsobs;  
	
	int i = 0, iring = 0, nphl = 0, ij = 0, smax = 0;
	int istart_south = 0, istart_north = 0; // map related 
	
	s2hat_flt8 dtmp = 0.0;  
	
	s2hat_mt_times* ct = NULL;
	
	int *indx = NULL;
	int *keep_it = NULL, *keep_it_all = NULL, *keep_north = NULL, *keep_south = NULL;
	
	s2hat_flt8 *ring = NULL, *kphi0 = NULL, *totWght = NULL, *nph0 = NULL;
	
	double complex *phas_n = NULL, *phas_s = NULL;
	fftw_complex *phring = NULL;
	
	// :: new stuff 
	
	int mfac_test = 0, mfac_lastm = 0;

	s2hat_flt8 mfac_two_x_m = 0.0, mfac_tmp_val = 0.0;
	s2hat_flt8* mfac = NULL;
	
	s2hat_flt8* rescale_tab = NULL;
	s2hat_flt8* local_alms_comp_a_real = NULL;
	s2hat_flt8* local_alms_comp_a_imag = NULL;
	
	struct fftw_counters_r2c fftwc = {0, 0, 0, 0, NULL};
	init_fft_r2c(&fftwc, nphmx/2);
	 
	// openMP stuff
	
	S2HATALLOC(indx, s2hat_int4, (nmmax + 1) );

	if (!(indx)) {
		perror("[mem error] Unable to allocate indx memory!\n");
		exit(-1);
	}
	
	fix_mmap(1, nmmax, indx, nmvals, mvals, myid, numprocs, comm);

	// --- allocates space for arrays
	
	// S2HATALLOC(ring, s2hat_flt8, nphmx);
	ring = (s2hat_flt8*) fftw_malloc(sizeof(s2hat_flt8) * (nphmx));
	phring = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * (nmmax + 1));
	// S2HATALLOC(phring, double complex, (nmmax + 1) );
	
	if (!(ring && phring && indx)) {
		perror("[mem error] Unable to allocate ring memory!\n");
		exit(-1);
	}
	
   const int nphas = max_((nmmax + 1)*nrings, nmvals*nringsobs);
   
	S2HATALLOC( phas_n, double complex, nphas );
	S2HATALLOC( phas_s, double complex, nphas );
	
	S2HATALLOC(keep_north, s2hat_int4, nrings);
	S2HATALLOC(keep_south, s2hat_int4, nrings);
	
	S2HATALLOC(keep_it, s2hat_int4, nrings);
	
	S2HATALLOC(nph0, s2hat_flt8, nrings);
	S2HATALLOC(kphi0, s2hat_flt8, nrings);
	S2HATALLOC(totWght, s2hat_flt8, nrings);
	
	if (!(nph0 && kphi0 && keep_north && keep_south && totWght)) {
		perror("[mem error] Unable to allocate phi memory!\n");
		exit(-1);
	}
	
	S2HATALLOC(local_alms_comp_a_real, s2hat_flt8, ((nlmax + 1) * nmvals) );
	S2HATALLOC(local_alms_comp_a_imag, s2hat_flt8, ((nlmax + 1) * nmvals) );
	
	if (!(local_alms_comp_a_real && local_alms_comp_a_imag)) perror("[mem error] Unable to allocate memory!\n");
	
	for (i = first_ring; i <= last_ring; ++i) {
		nph0 [i - first_ring] = (s2hat_flt8) pixelization.nph[i - 1];
		kphi0[i - first_ring] = (s2hat_flt8) pixelization.kphi[i - 1];
		totWght[i -first_ring] = pixelization.parea[i -1]*pixelization.qwght[i -1];
		
		// find out which rings are to be analysed
		
		keep_north[i - first_ring] = (s2hat_int4) scan.nfl[i - 1];
		keep_south[i - first_ring] = (s2hat_int4) scan.sfl[i - 1];
		keep_it[i - first_ring] = (s2hat_int4) scan.fl[i - 1];
	}
	
	istart_north = 0;
	istart_south = map_size;
	
	if (ct != NULL) fft_start = MPI_Wtime();

	for (iring = 0; iring < nrings; ++iring) {
		nphl = nph0[iring];
		
		if (keep_north[iring] == 1) {
			
			memcpy(ring, local_map + istart_north, nphl * sizeof (s2hat_flt8));
			dtmp = local_w8ring[iring] * totWght[iring];
			
			for (i = 0; i < nphl; ++i) {
				ring[i] *= dtmp;
			}
						
			s2hat_ring_analysis(nlmax, nmmax, ring, nphl, phring, kphi0[iring], &fftwc); // north hemisph. + equator
			
			for (i = 0; i <= nmmax; ++i) {
				ij = iring + indx[i] * nrings;
				phas_n[ij] = (double complex) phring[i];
			}
		}

		istart_north = istart_north + nphl;
		istart_south = istart_south - nphl;

		if (keep_south[iring] == 1) {

			memcpy(ring, local_map + istart_south, nphl * sizeof (s2hat_flt8));
			
			
			dtmp = local_w8ring[iring] * totWght[iring];
			
			for (i = 0; i < nphl; ++i) {
				ring[i] *= dtmp;
			}
			
			s2hat_ring_analysis(nlmax, nmmax, ring, nphl, phring, kphi0[iring], &fftwc); // south hemisph. w/o equat
			
			for (i = 0; i <= nmmax; ++i) {
				ij = iring + indx[i] * nrings;
				phas_s[ij] = (double complex) phring[i];
			}
		}
	}
	
	if (ct != NULL) fft_end = MPI_Wtime();
	
	if (ct != NULL) communication_start = MPI_Wtime();
			
	redistInterProductsMap2Alm(nrings, nmvals, nphas, phas_n, myid, numprocs, comm);	
	redistInterProductsMap2Alm(nrings, nmvals, nphas, phas_s, myid, numprocs, comm);
	
	if (ct != NULL) communication_end = MPI_Wtime();
	
	S2HATALLOC(keep_it_all, s2hat_int4, nringsall);
	
	gatherRingInfo(first_ring, nrings, keep_it, nringsall, keep_it_all, myid, numprocs, comm);
	s2hat_free(keep_it);
	keep_it = NULL;
	
	// Generate mfac proper //

	S2HATALLOC(mfac, s2hat_flt8, (nmmax + 1) );
	
	mfac_test = (mvals[0] == 0) ? 1 : 0;

	if (mfac_test) {
		mfac[0] = 1.0;
	}

	mfac_tmp_val = 1.0;
	mfac_lastm = 1;

	for (i = (mfac_test) ? 1 : 0; i < nmvals; ++i) {
		while (mfac_lastm <= mvals[i]) {
			mfac_two_x_m = 2.0 * mfac_lastm;
			mfac_tmp_val = mfac_tmp_val * sqrt((mfac_two_x_m + 1.0) / mfac_two_x_m);
			++mfac_lastm;
		}
		mfac[i] = mfac_tmp_val;
	}

	for (i = 0; i < nmvals; ++i) {
		mfac[i] = log(SQ4PI_INV * mfac[i]) * ALN2_INV;
	}
	
	smax = init_rescale_local(&rescale_tab); // initialize rescale table

	if (smax == 0) {
		perror("[mem error] Unable to init rescale table!\n");
	}
   	
	s2hat_flt8 ovflow = RESCALE_TAB(1);
	s2hat_flt8 unflow = RESCALE_TAB(-1);
		
	if (ct != NULL) recurence_start = MPI_Wtime();
	
#pragma omp parallel shared(mvals) shared(nlmax) shared(pixelization) shared(phas_n) shared(phas_s) shared(local_alm) shared(nmmax) shared(mfac) shared(rescale_tab) shared(myid)
{  
	int mindx = 0, jring = 0, m = 0, l = 0, lalm_shift = 0;
	int ithl = 0, scalem = 0, scalel = 0;
	
	int l_min = 0, par_lm = 0;
	
	s2hat_flt8 lam_mm = 0.0, lam_lm = 0.0, lam_0 = 0.0, lam_1 = 0.0, lam_2 = 0.0, corfac = 0.0, cth_ring = 0.0;
	
	double complex php, phm;
	
	ComplexStructPtr recfac_0 = {NULL, NULL}, recfac_1 = {NULL, NULL};
	
	const int rank_hyb   = omp_get_thread_num(); 
	const int nbproc_hyb = omp_get_num_threads();
	
	const int size_tmp = nmvals / (nbproc_hyb * 2);
	const int remains  = (rank_hyb == 0) ? nmvals % size_tmp : 0;
	
	const int begin0 = rank_hyb * size_tmp;
	const int end0 = begin0 + size_tmp;
	
	const int end1 = (((2 * nbproc_hyb) - rank_hyb) * size_tmp) + remains;
	const int begin1 = end1 - remains - size_tmp;
		
	for (mindx = begin0; mindx < end0; ++mindx) {
		m = mvals[ mindx ];
		lalm_shift = (mindx)*(nlmax + 1);        
		gen_recfac(nlmax, m, &recfac_0);
		
		double* dalm_re = local_alms_comp_a_real + lalm_shift; //skip mindx L columns
		double* dalm_im = local_alms_comp_a_imag + lalm_shift; //skip mindx L columns
		
		for (i = m; i <= nlmax; ++i) {
			dalm_re[i] = 0.0;
			dalm_im[i] = 0.0;
		}
		
		jring = 0;
		for (ithl = 0; ithl < nringsall; ++ithl) { //uchk-lchk
			if (keep_it_all[ithl] == 1) {

				l_min = l_min_ylm(m, pixelization.sth[ithl]);

				if (nlmax >= l_min) {
					
					mpi_compute_lam_mm(mfac[mindx], m, pixelization.sth[ithl], smax, rescale_tab, &lam_mm, &corfac, &scalem);
					
					// ===================
					// =		 l = m		=
					// ===================
					
					par_lm = 1;
					 			
					php = phas_n[mindx*nringsobs+jring] + phas_s[mindx*nringsobs+jring]; // sum  (if (l+m) even)
					phm = phas_n[mindx*nringsobs+jring] - phas_s[mindx*nringsobs+jring]; // diff (if (l+m) odd)
					
					corfac = corfac*lam_mm;
					
					if (m >= l_min) { // = (-1)^(l+m)
						lam_lm = corfac;
						
						dalm_re[m] += (lam_lm * ((par_lm > 0) ? creal(php) : creal(phm)));
						dalm_im[m] += (lam_lm * ((par_lm > 0) ? cimag(php) : cimag(phm)));
						
					}
					
					// ===================
					// =		 l > m		=
					// ===================

					lam_0 = 0.0;
					lam_1 = 1.0;
					scalel = scalem;
					cth_ring = pixelization.cth[ithl];
					lam_2 = cth_ring * lam_1 * recfac_0.re[m];


					for (l = m + 1; l <= nlmax; ++l) {
						
						par_lm = -par_lm;  // = (-1)^(l+m)  
						
						if (l >= l_min) {
							lam_lm = lam_2 * corfac;
							
							dalm_re[l] += (lam_lm * ((par_lm > 0) ? creal(php) : creal(phm)));
							dalm_im[l] += (lam_lm * ((par_lm > 0) ? cimag(php) : cimag(phm)));
							
						}
						
						lam_0 = lam_1 * recfac_0.im[l - 1];
						lam_1 = lam_2;
						lam_2 = (cth_ring * lam_1 - lam_0) * recfac_0.re[l];

						if (absolute(lam_2) > ovflow) {

							lam_1 = lam_1*unflow;
							lam_2 = lam_2*unflow;
							scalel = scalel + 1;

							corfac = lam_mm * RESCALE_TAB(max_(scalel, -smax));
							
						} else if (absolute(lam_2) < unflow) {
							lam_1 = lam_1*ovflow;
							lam_2 = lam_2*ovflow;
							scalel = scalel - 1;

							corfac = lam_mm * RESCALE_TAB(max_(scalel, -smax));
						}
					}//~end FOR loop on l
				}//~end test for L vals
				++jring;
			}
		}//~end loop over rings
		for ( l = m; l <= nlmax; ++l) {
			local_alm[lalm_shift + l].re += dalm_re[l];
			local_alm[lalm_shift + l].im += dalm_im[l];
		}
	}//~end loop over m
	
	s2hat_free(recfac_0.re);
   s2hat_free(recfac_0.im);

	for (mindx = begin1; mindx < end1; ++mindx) {
		m = mvals[ mindx ];
		lalm_shift = (mindx)*(nlmax + 1);
		gen_recfac(nlmax, m, &recfac_1);
		
		double* dalm_re = local_alms_comp_a_real + lalm_shift; //skip mindx L columns
		double* dalm_im = local_alms_comp_a_imag + lalm_shift; //skip mindx L columns
		
		for (i = m; i <= nlmax; ++i) {
			dalm_re[i] = 0.0;
			dalm_im[i] = 0.0;
		}
		
		jring = 0;
		for (ithl = 0; ithl < nringsall; ++ithl) { //uchk-lchk
			
			l_min = l_min_ylm(m, pixelization.sth[ithl]);
			
			if (keep_it_all[ithl] == 1) {
				if (nlmax >= l_min) {
					
					mpi_compute_lam_mm(mfac[mindx], m, pixelization.sth[ithl], smax, rescale_tab, &lam_mm, &corfac, &scalem);
					
					// ===================
					// =		 l = m		=
					// ===================
					
					par_lm = 1;
					
					php = phas_n[mindx*nringsobs+jring] + phas_s[mindx*nringsobs+jring]; // sum  (if (l+m) even)
					phm = phas_n[mindx*nringsobs+jring] - phas_s[mindx*nringsobs+jring]; // diff (if (l+m) odd)
					
					corfac = corfac*lam_mm;
					
					if (m >= l_min) { // = (-1)^(l+m)
						lam_lm = corfac;
						
						dalm_re[m] += (lam_lm * ((par_lm > 0) ? creal(php) : creal(phm)));
						dalm_im[m] += (lam_lm * ((par_lm > 0) ? cimag(php) : cimag(phm)));
						
					}
					
					// ===================
					// =		 l > m		=
					// ===================
	
					lam_0 = 0.0;
					lam_1 = 1.0;
					scalel = scalem;
					cth_ring = pixelization.cth[ithl];
					lam_2 = cth_ring * lam_1 * recfac_1.re[m];
	
	
					for (l = m + 1; l <= nlmax; ++l) {
						
						par_lm = -par_lm;  // = (-1)^(l+m)  
						
						if (l >= l_min) {
							lam_lm = lam_2 * corfac;
							
							dalm_re[l] += (lam_lm * ((par_lm > 0) ? creal(php) : creal(phm)));
							dalm_im[l] += (lam_lm * ((par_lm > 0) ? cimag(php) : cimag(phm)));
							
						}
						
						lam_0 = lam_1 * recfac_1.im[l - 1];
						lam_1 = lam_2;
						lam_2 = (cth_ring * lam_1 - lam_0) * recfac_1.re[l];
	
						if (absolute(lam_2) > ovflow) {
	
							lam_1 = lam_1*unflow;
							lam_2 = lam_2*unflow;
							scalel = scalel + 1;
	
							corfac = lam_mm * RESCALE_TAB(max_(scalel, -smax));
							
						} else if (absolute(lam_2) < unflow) {
							lam_1 = lam_1*ovflow;
							lam_2 = lam_2*ovflow;
							scalel = scalel - 1;
	
							corfac = lam_mm * RESCALE_TAB(max_(scalel, -smax));
						}
					}//~end FOR loop on l
				}//~end test for L vals
				++jring;
			}
		}//~end loop over rings
		for ( l = m; l <= nlmax; ++l) {
			local_alm[lalm_shift + l].re += dalm_re[l];
			local_alm[lalm_shift + l].im += dalm_im[l];
		}
	}//~end loop over m
		
	s2hat_free(recfac_1.re);
	s2hat_free(recfac_1.im);
		  
}//openMP section

	if (ct != NULL) recurence_end = MPI_Wtime();
	
	if (ct != NULL) {
		
		ct->recurrence_dt = recurence_end - recurence_start;
		ct->fft_dt        = fft_end - fft_start;
		ct->comm_dt       = communication_end - communication_start;
		
	}

	free_fft_r2c(&fftwc);
	
	s2hat_free(mfac);
	s2hat_free(rescale_tab);
	
	s2hat_free(keep_it_all);
	s2hat_free(kphi0);
	s2hat_free(nph0);
	s2hat_free(keep_south);
	s2hat_free(keep_north);	
	s2hat_free(phas_s);
	s2hat_free(phas_n);
	fftw_free(phring);
	fftw_free(ring);
	s2hat_free(indx); 
}

